#!/usr/local/bin/perl -w
#
# cgishell.pl
#
# A CGI script which simulates a command-line interface. Use it on a
# server where you have CGI access but wouldn't otherwise have telnet
# access (e.g. a commercial web host), or if you can't telnet to the box
# for any other reason.
#
# See http://www.gloomytrousers.co.uk/util/ for latest version, more
# detailed instructions, and contact details.
#
# Copyright (c) R. Odom 2002
#
#	This program is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; either version 2 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful, but
#	WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#	General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#	USA
#
# Installation:
#	Just drop it in a directory that can execute CGI scripts. Make it
#	executable for the appropriate user (depends on web server). Set up
#	authentication on the script using .htaccess or whatever.
#
# Notes:
#	* For security, the script will exit (with a suitable error message)
#	  if the remote user has not authenticated with the server (i.e. if
#	  AUTH_USER is not set).
#	* Only tested (so far) under IIS [ToDo: this will be rectified in a
#	  few weeks].
#	* In M$ Internet Exploiter, there is a bug (no shit, Sherlock!) -
#	  when you set the innerHTML property, newlines, tabs, etc are
#	  converted to spaces, even on a <PRE>. I've partially fudged it by
#	  converting spaces to non-breaking spaces if IE is detected, but
#	  tabs will appear incorrectly (as single spaces). Get a proper
#	  browser if this bothers you.
#	* The command 'cd' is handled as a special case. Any other command
#	  which changes the working directory won't have the desired effect
#	  - I don't know a way of telling the working directory after the
#	  command exits [ToDo]. 'cd -' (on operating systems that would
#	  normally support it) won't work [ToDo: maintain directory
#	  history?]. On Windows, 'cd /d [drive]' doesn't work either; to
#	  change to a different drive just type 'D:\' (where D is the drive
#	  you want). The slash is required.
#	* HTTP is stateless! This means that the environment won't be
#	  preserved between commands. [ToDo - I can possibly do this by
#	  comparing initial and final environments, and resolving any
#	  discrepancies before executing the command]
#
# Version history:
#	v0.1	31 Aug 2002
#		Initial version. Quite a few ToDos still left. Beta quality
#		only!
#	v0.1a	09 Sep 2002
#		Added GPL. Minor improvements to comments/instructions above.
#

$VER = '0.1a';

use CGI;
use Cwd;

# For security reasons, we don't want just any muppet to be able to run this
# Therefore, we return an error unless they've authenticated with the server
# Use .htaccess or similar to control this, or ask your hosting provider to set it up, or something
if (!defined($ENV{'AUTH_USER'}) or !$ENV{'AUTH_USER'}) {
	&error("You must be an authorised user on this resource to access this script.");
}

# Work out who we are running as
foreach $try ($ENV{'USER'}, $ENV{'USERNAME'}, `whoami`, `who am i`, '[unknown]') {
	if (defined($try) && $try) {
		chomp($try);
		$user = $try;
		last;
	}
}

# See if we can find a temporary directory that we can write to
foreach $try ($ENV{'TEMP'}, $ENV{'TMP'}, '/var/tmp', 'C:/') {
	chomp($try);
	if (defined($try) && $try && -w $try) {
		$tempname = $try;
		last;
	}
}
$tempname .= '/.cgishell$$.tmp';	### ToDo: under Windows, if we change drives and didn't get a temp dir (unlikely) this path may become invalid

# Attempt to grab STDERR, just to check that we can (some servers lock it)
open(STDERR_SAVE,">&STDERR");
$stderr = open(STDERR, ">$tempname");
unlink("$tempname");
open(STDERR,">&STDERR_SAVE");

# Work out what to do
$q = new CGI;
if (!defined($q->param('action'))) {
	&mainscreen();
} elsif ($q->param('action') eq 'start') {
	&shellscreen();
} elsif ($q->param('action') eq 'env') {
	&env();
} elsif ($q->param('action') eq 'runcmd') {
	&runcmd($q->param('cmd'), $q->param('dir'));
} else {
	&error('Invalid action');
}


# The main screen, includes all the frames and stuff
sub mainscreen() {
	print <<ENDHTML;
Content-type: text/html

<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en-gb" lang="en-gb">
<head>
	<title>cgishell: $user\@$ENV{'SERVER_NAME'}</title>
	<style type="text/css">
	.error {
		color: red;
	}
	</style>
	<script type="text/javascript">
	var commandActive = false,
		cmd = '';

	// See if there's a command running already
	function checkOKToSend() {
		if (commandActive) {
			// Could get into a race condition with commandDone
			if (! confirm('A command is currently active; you can submit your new command, but the current command will continue to run and you will not see its output. Click OK to continue waiting for the current command, or Cancel to ignore it and send your new one.')) {
					return false;
				} else {
					addCommandToHistory (document.forms['prompt'].cmd.value);
					commandActive = true;
					return true;
				}
		} else {
			addCommandToHistory (document.forms['prompt'].cmd.value);
			commandActive = true;
			return true;
		}
	}

	// Add a command to the history select box
	function addCommandToHistory (commandtext) {
		if (commandtext != '') {
			// ToDo: make this work top-down? Alternatively, scroll to the bottom of the select box?
			document.forms['util'].commandhistory.options[document.forms['util'].commandhistory.options.length] =
				new Option(String(commandtext).substr(0, 40) + (String(commandtext).length > 40?'...':''),
							commandtext,
							false,
							false);
			document.forms['util'].commandhistory.selectedIndex = document.forms['util'].commandhistory.options.length - 1;
		}
	}

	// Called by the hidden command frame once the command has finished
	function commandDone() {
		// Set the new working dir, if available
		if (window.frames['shell'].document.forms['results'].newdir.value != '') {
			document.forms['prompt'].dir.value = window.frames['shell'].document.forms['results'].newdir.value;
		}

		// Add the results to our 'screen'
		window.frames['virtualscreen'].addResult(
			window.frames['shell'].document.forms['results'].cmd.value,
			window.frames['shell'].document.forms['results'].stdout.value,
			window.frames['shell'].document.forms['results'].stderr.value,
			window.frames['shell'].document.forms['results'].newdir.value
		);

		// Prepare for next command
		document.forms['prompt'].cmd.value = '';
		document.forms['prompt'].cmd.focus();
		commandActive = false;
	}

	</script>
</head>

<body onload="window.frames['virtualscreen'].location.href = 'cgishell.pl?action=start'; document.forms['prompt'].cmd.focus()">
	<div style="position: absolute; top: 0; left: 10px; width: 60%">
		<h1>cgishell</h1>
		<p>This is <a href="http://www.gloomytrousers.co.uk/util/#cgishell" target="_blank"><code>cgishell</code></a> version $VER, running as <b>$user\@$ENV{'SERVER_NAME'}</b></p>

		<noscript><div class="error">Error: <code>cgishell</code> needs JavaScript enabled to work.</div></noscript>
ENDHTML
	if (!$stderr) {
		print '<div class="error">Warning: Can\'t grab STDERR (either the server locks it, or I haven\'t got permission to write a temporary file) - if your commands output anything to STDERR, you won\'t see it!</DIV>';
	}
	print<<ENDHTML;
		<iframe name="virtualscreen" src="about:blank" style="width: 100%; height: 25em;">
			<div class="error">Error: <code>cgishell</code> needs a browser which supports IFRAMEs.</div>
		</iframe><br /><br />

		<form name="prompt" action="cgishell.pl" method="post" target="shell" onsubmit="return checkOKToSend()">
			<input type="hidden" name="action" value="runcmd">
			Working dir: <input type="text" name="dir" size="40" value=""><br />
			Command: <input type="text" name="cmd" size="40" value="">
			<script type="text/javascript">
				// Sometimes a reload causes this to get displayed in the history window; clear it ASAP to prevent this
				document.forms['prompt'].cmd.value = '';
			</script>
			<input type="submit" value="Execute">
		</form>
	</div>

	<form name="util" style="position: absolute; top: 0; right: 10px; width: 30%">
		<h2>Environment</h2>
		<iframe name="env" src="cgishell.pl?action=env" style="width: 100%; height: 10em;"></iframe>
		<input type="button" value="Refresh" onClick="window.frames['env'].location.reload(false)"><br />

		<h2>Command history</h2>
		<select name="commandhistory" size="5" onclick="document.forms['prompt'].cmd.value = this.options[this.selectedIndex].value; document.forms['prompt'].cmd.focus()">
		</select>

		<!-- this is the frame where the command gets executed; you don't need to see it -->
		<iframe name="shell" src="about:blank" style="width: 1px; height: 1px; display: none;"></iframe>
	</form>
</body>
</html>
ENDHTML
}

# The virtual screen, where the command results appear
sub shellscreen() {
	print <<ENDHTML;
Content-type: text/html

<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en-gb" lang="en-gb">
<head>
	<title>Pseudo-screen</title>
	<style type="text/css">
	body {
		background-color: white;
		margin: 0;
	}
	#text {
		padding: 0.2em;
		margin: 0;
	}
	.cmd {
		color: navy;
	}
	.dir {
		color: green;
	}
	.err {
		color: maroon;
	}
	.out {
		color: black;
	}
	#readymsg {
		display: none;
	}
	</style>
	<script type="text/javascript">
	var firstCommand = true;

	// Add a command and its result to the 'screen'
	function addResult(cmd, stdout, stderr, newdir) {

		// Show the fact that we're ready if this is the first command
		if (firstCommand) {
			document.getElementById('readymsg').style.display = 'inline';
		}

		// ToDo: Manually wrap output at 80 chars?

		// Encode ready for output
		cmd = HTMLEncode(cmd);
		stdout = HTMLEncode(stdout);
		stderr = HTMLEncode(stderr);
		newdir = HTMLEncode(newdir) + '\$ ';	// Add prompt char to the end of this

		// IE bug: ignores newlines when you set innerHTML, even on a PRE.
		// This is a fudge; the output won't be right if there's tabs in it. I hate having to do this kind of thing.
		if (String(navigator.appName).indexOf('Internet Explorer') > -1) {
			if (cmd != '') {
				cmd = cmd.replace(/ /g, '&nbsp;');
				cmd = cmd.replace(/\\n/g, '<br />') + '<br />';
			}
			if (stdout != '') {
				stdout = stdout.replace(/ /g, '&nbsp;');
				stdout = stdout.replace(/\\n/g, '<br />' + '<br />');
			}
			if (stderr != '') {
				stderr = stderr.replace(/ /g, '&nbsp;');
				stderr = stderr.replace(/\\n/g, '<br />') + '<br />';
			}
			if (newdir != '') {
				newdir = newdir.replace(/ /g, '&nbsp;');
				newdir = newdir.replace(/\\n/g, '<br />');
			}
			//alert (stdout);
		}

		var text = new String((cmd?'<span class="cmd">' + cmd + '</span>':'') + '\\n' +
					(stdout?'<span class="out">' + stdout + '</span>\\n':'') +
					(stderr?'<span class="err">' + stderr + '</span>\\n':'') +
					'<span class="dir">' + newdir + '</span>');

		// Write to our screen
		document.getElementById('text').innerHTML = document.getElementById('text').innerHTML + text;

		// Scroll to bottom
		window.scrollBy(0, 1000000);
	}

	function HTMLEncode(val) {
		// ToDo: Need to strip HTML better
		val = String(val).replace(/&/g, '&amp;');
		val = val.replace(/"/g, '&quot;');	//"
		val = val.replace(/</g, '&lt;');
		val = val.replace(/>/g, '&gt;');

		return val;
	}
	</script>
</head>
<body onload="parent.frames['shell'].location.href = 'cgishell.pl?action=runcmd'">
<pre id="text">cgishell loading...<span id="readymsg"> ready!</span></pre>
</body>
</html>
ENDHTML
}


# Runs the command and returns the results as hidden form fields for the front end to process
sub runcmd () {
	my $cmd = $_[0];
	my $dir = $_[1];
	my $result = '';
	my $errors = '';

	# Change to the correct dir for this command
	if (defined($dir) && $dir) {
		chdir($dir) or &error("Can't cd to $dir", "Your command has not been executed");
	} else {
		$dir = '.';
	}

	# This is a workaround for now, until I work out how to tell what directory the command leaves us in
	if ($cmd =~ /^cd\s+(.+)$/i and chdir($1)) {
		# chdir was successful; don't need to execute the command
		# (If the chdir fails, we drop through and execute the command, which gives us an error message for the user)
	} elsif (defined($cmd) && $cmd) {

		# Trap STDERR before we do the command?
		if ($stderr) {
			open(STDERR_SAVE,">&STDERR");
			open(STDERR, ">$tempname") or &error('Couldn\'t open STDERR', 'Your command has not been executed');
		}

		# Do the command
		$result = `$cmd`;	### ToDo: We need to know if this changes the working directory

		if ($stderr) {
			# Restore STDERR to original
			open(STDERR,">&STDERR_SAVE");
			# Slurp up our error log into a variable
			if (open(ERR, "$tempname")) {
				while(<ERR>) {
					$errors .= $_;
				}
				close(ERR);
			} else {
				$errors = 'Your command may have produced output on STDERR, but I was unable to open the file containing it';
			}
			# Get rid of the log now we're done with it
			unlink("$tempname");
		}
	}

	# Find our working dir now, in case it's changed
	my $newdir = getcwd();	### ToDo: This doesn't work, since the command starts a new shell above
	chomp $newdir;

	# Encode ready for output
	$cmd = &HTMLEncode($cmd);
	$newdir = &HTMLEncode($newdir);
	$result = &HTMLEncode($result);
	$errors = &HTMLEncode($errors);

	print <<ENDHTML;
Content-type: text/html

<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en-gb" lang="en-gb">
<head>
	<title>$cmd</title>
	<style type="text/css">
	.errors {
		color: red;
	}
	</style>
</head>
<body onload="parent.commandDone()">
<form name="results">
	<input type="hidden" name="cmd" value="$cmd">
	<input type="hidden" name="newdir" value="$newdir">
	<input type="hidden" name="stdout" value="$result">
	<input type="hidden" name="stderr" value="$errors">
</form>
<!-- You don't need to see this frame -->
</body>
</html>
ENDHTML
}


# Print a list of all the environment variables (how useful is this?)
sub env() {
	print <<ENDHTML;
Content-type: text/html

<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en-gb" lang="en-gb">
<head>
	<title>Environment</title>
</head>
<body style="margin: 0; padding: 0;">
<pre style="padding: 0.2em; margin: 0;">
ENDHTML
	foreach $key (sort keys %ENV) {
		print "<b>$key</b>=";
		# Hide sensitive info; user can still get at it if they do 'set', but we don't want to show it by default
		if ($key =~ /^(HTTP_AUTHORIZATION|AUTH_PASSWORD)$/i) {
			print "<i>[hidden]</i>\n";
		} else {
			print HTMLEncode($ENV{$key}) . "\n";
		}
	}
	print <<ENDHTML;
</pre>
</body>
</html>
ENDHTML
}


# Returns an error message (HTML) then exits
sub error() {
	print<<ENDHTML;
Content-type: text/html

<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en-gb" lang="en-gb">
<head>
	<title>cgishell: Error</title>
</head>
<body onload="if (typeof(parent.commandDone) != 'undefined') { parent.commandDone() } ">
	<h1>Error</h1>
	<ul>
ENDHTML
	foreach $_ (@_) {
		print "\t\t<li>$_</li>\n";
	}
	print<<ENDHTML;
	</ul>
	<form name="results">
		<input type="hidden" name="cmd" value="">
		<input type="hidden" name="newdir" value="">
		<input type="hidden" name="stdout" value="*** cgishell internal error: ***">
		<input type="hidden" name="stderr" value="
ENDHTML
	foreach $_ (@_) {
		print "$_\n";
	}
	print<<ENDHTML;
		">
	</form>
</body>
</html>
ENDHTML
	die(@_);
}


# Escape characters that are special in HTML
sub HTMLEncode() {
	my $val = $_[0];
	# ToDo: Need to strip HTML better; this does the trick if the browser is forgiving (most of 'em!) but needs to be better really
	$val =~ s/\&/\&amp;/g;
	$val =~ s/"/\&quot;/g; # "
	$val =~ s/</\&lt;/g;
	$val =~ s/>/\&gt;/g;

	return $val;
}

__END__